
assert( LibStub, "LibStub is not loaded." )

local addon = LibStub( "AceAddon-3.0" ):NewAddon( "hcGossipBot", "AceConsole-3.0", "AceEvent-3.0" )

local DEFAULTS = {
	profile = {
		enabled = true,
	},
}

function addon:OnInitialize( )
	self.db = LibStub( "AceDB-3.0" ):New( self.name .. "SavedVariables", DEFAULTS, true )
	self:SetEnabledState( self.db.profile.enabled )
end

local EVENTS = {
	"GOSSIP_SHOW",
	"QUEST_ACCEPT_CONFIRM",
	"QUEST_AUTOCOMPLETE",
	"QUEST_COMPLETE",
	"QUEST_DETAIL",
	"QUEST_PROGRESS",
}

function addon:OnEnable( )
	self.db.profile.enabled = true
	for _, event in ipairs( EVENTS ) do
		self:RegisterEvent( event )
	end
end

function addon:OnDisable( )
	self.db.profile.enabled = false
end

--- Tests whether or not interaction with the current NPC is disabled.
-- If the SHIFT key is pressed, then automation is enabled.
-- @return true if automatic interaction should be disabled.
function addon:IsGossipHandlerDisabled( )
	return not IsShiftKeyDown( )
end

function addon:IsChoiceHandlerDisabled( )
	return not IsControlKeyDown( )
end

-- TODO write me
local function _FindCompleteQuest( ... )
	local index = 1
	for i = 1, select( "#", ... ), 4 do
		local isComplete = select( i + 3, ... )
		if isComplete then
			return index
		end
		index = index + 1
	end
	
	return 0
end

-- Returns true if the player should accept a quest.
-- This returns false if automatic accepting of quests is disabled, if
-- the quest log is full, or if the maximum number of daily quests has 
-- already been completed while checking a daily quest.
-- @param true if the quest being checked is a daily quest.
local function _ShouldAcceptQuest( isDaily )
	local _, numQuests = GetNumQuestLogEntries( )
	if numQuests >= MAX_QUESTS then
		return false
	end
	
	if isDaily and ( GetDailyQuestsCompleted( ) >= GetMaxDailyQuests( ) ) then
		return false
	end
	
	return true
end

-- TODO write me
local function _FindAcceptQuest( ... )
	local index = 1
	for i = 1, select( "#", ... ), 5 do
		local isDaily = select( i + 3, ... )
		if _ShouldAcceptQuest( isDaily ) then
			return index
		end
		index = index + 1
	end
	
	return 0
end

local function _IsNpcMerchant( )
	return GetMerchantNumItems( ) > 0 or CanMerchantRepair( )
end

--- Fires when an NPC gossip interaction begins.
-- This is fired for most initial NPC interactions. 
function addon:GOSSIP_SHOW( event )
	if self:IsGossipHandlerDisabled( ) then
		return
	end
	
	-- If any active quests have been completed, turn in the first one 
	-- the active quest list.
	if GetNumGossipActiveQuests( ) > 0 then
		local index = _FindCompleteQuest( GetGossipActiveQuests( ) )
		if index > 0 then
			SelectGossipActiveQuest( index )
			return
		end
	end
	
	-- If any quests are available, accept the first one in the available
	-- quest list.
	if GetNumGossipAvailableQuests( ) > 0 then
		local index = _FindAcceptQuest( GetGossipAvailableQuests( ) )
		if index > 0 then
			SelectGossipAvailableQuest( index )
			return
		end
	end

	-- If only one gossip option is available, select it.
	if GetNumGossipOptions( ) == 1 then
		self:PrintGossip( true, ( GetGossipOptions( ) ) )
		self:PrintGossip( false, GetGossipText( ) )
		SelectGossipOption( 1 )
		return
	end
	
	if ( GetNumGossipOptions( ) + GetNumGossipAvailableQuests( ) + GetNumActiveQuests( ) == 0 ) then
		-- This seems to fire for merchants
		if not _IsNpcMerchant( ) then
			self:PrintGossip( false, GetGossipText( ) )
		end
		CloseGossip( )
		return
	end
end

---	Fires when escort quests are started by a party member.
-- @param name the name of the user who started the quest.
-- @param quest te name of the quest that was started.
function addon:QUEST_ACCEPT_CONFIRM( event, name, quest )
	if self:IsGossipHandlerDisabled( ) then
		return
	end
	
	if _ShouldAcceptQuest( ) then
		ConfirmAcceptQuest( )
	end
end

---	Fires when a quest is automatically completed.
-- These types of quests may be handed in remotely, and will pop up 
-- with a completion indicator in the objective list.
-- @param questId the id of the quest completed.
function addon:QUEST_AUTOCOMPLETE( event, questId )
	if self:IsGossipHandlerDisabled( ) then
		return
	end
	
	ShowQuestComplete( questId )
end

-- Returns the vendor price of a given item.
-- @param item an item id, item name, or item link
-- @return the vendor price, in copper.
local function _GetItemVendorPrice( item )
	local _, _, _, _, _, _, _, _, _, _, vendorPrice = GetItemInfo( item )
	return vendorPrice
end

-- Returns the quest item choice with the highest vendor value.
-- @return the choice index of the highest vendor value item.
local function _GetBestQuestItemChoice( )
	local bestChoice = nil
	local bestPrice = nil
	for choice = 1, GetNumQuestChoices( ) do
		local item = GetQuestItemLink( "choice", choice )
		local price = _GetItemVendorPrice( item )
		if ( not bestPrice ) or ( bestPrice < price ) then
			bestChoice = choice
			bestPrice = price
		end
	end
	return bestChoice
end

---	Fires when the player is looking at the Complete page for a quest.
-- This may be at a questgiver, or may be via remote completion.
function addon:QUEST_COMPLETE( event )
	if self:IsGossipHandlerDisabled( ) then
		return
	end
	
	local choice = nil
	
	if GetNumQuestChoices( ) > 0 then
		if self:IsChoiceHandlerDisabled( ) then
			return
		end
		choice = _GetBestQuestItemChoice( )
	end
	
	self:PrintGossip( false, GetProgressText( ) )
	GetQuestReward( choice )
end

--- Fires when details of an available quest are presented by a questgiver.
function addon:QUEST_DETAIL( event )
	if self:IsGossipHandlerDisabled( ) then
		return
	end
	
	if not _ShouldAcceptQuest( QuestIsDaily( ) ) then
		DeclineQuest( )
		return
	end

	self:PrintGossip( false, GetQuestText( ), GetObjectiveText( ) )
	AcceptQuest( )
end

--- Fires when interacting with a questgiver about an active quest.
function addon:QUEST_PROGRESS( event )
	if self:IsGossipHandlerDisabled( ) then
		return
	end
	
	self:PrintGossip( false, GetProgressText( ) )
	
	CompleteQuest( )
end

-- TODO document me
local function _GetMessageTypeColorHex( messageType )
	local r, g, b, _ = GetMessageTypeColor( messageType )
	return string.format( "|cff%02x%02x%02x", r * 255, g * 255, b * 255 )
end

-- TODO document me
local function _GetClassTypeColorHex( classType )
	local color = RAID_CLASS_COLORS[ classType ]
	return string.format( "|cff%02x%02x%02x", color.r * 255, color.g * 255, color.b * 255 )
end

-- TODO document me
local function _CreateNpcEmotePrefix( )
	local textColor = _GetMessageTypeColorHex( "CHAT_MSG_EMOTE" )
	return textColor
end

-- TODO document me
local function _CreateEmotePrefix( player )
	return _CreateNpcEmotePrefix( )
end

-- TODO document me
local function _CreateNpcGossipPrefix( )
	local textColor = _GetMessageTypeColorHex( "CHAT_MSG_MONSTER_SAY" )
	local guid = UnitGUID( "npc" )
	local name = UnitName( "npc" )
	return string.format( "%s|Hunit:%s:%s|h%s|h says: ", textColor, guid, name, name )
end

-- TODO document me
local function _CreatePlayerGossipPrefix( )
	local textColor, _ = _GetMessageTypeColorHex( "CHAT_MSG_SAY" )
	local name = UnitName( "player" )
	local _, classType = UnitClass( "player" )
	local classColor = _GetClassTypeColorHex( classType )
	return string.format( "%s|Hplayer:%s|h[%s]|h%s says: ", classColor, name, name, textColor )
end

-- TODO document me
local function _CreateGossipPrefix( player )
	if player then
		return _CreatePlayerGossipPrefix( )
	else
		return _CreateNpcGossipPrefix( )
	end
end

-- TODO document me
local function _PrintGossipItems( player, frame, ... )
	local emotePrefix = _CreateNpcEmotePrefix( )
	local gossipPrefix = _CreateGossipPrefix( player )
  for i = 1, select( '#', ... ) do
		local message = strtrim( select( i, ... ) )
		if strlen( message ) > 0 then
			local emote = string.match( message, "^<(.+)>$" )
			if emote then
				frame:AddMessage( emotePrefix .. emote )
			else
				frame:AddMessage( gossipPrefix .. message )
			end
		end
	end
end

--- Prints gossip text to the default chat frame.
-- If the gossip text contains multiple lines, each line will be output
-- individually. Each line of text is formatted to appear like normal 
-- creature speech messages.
-- @param player true if the gossip text is spoken by the player.
-- @param ... strings to be written.
function addon:PrintGossip( player, ... )
	for i = 1, select( "#", ... ) do
		local message = select( i, ... )
		if type( message ) == "string" then
			_PrintGossipItems( player, DEFAULT_CHAT_FRAME, strsplit( "\n", message ) )
		end
	end
end
